@@ -41,7 +41,7 @@
 #include "egg/egg-error.h"
 #include "egg/egg-unix-credentials.h"
 
-#include "gck/gck.h"
+#include <gck/gck.h>
 
 #include "pkcs11/pkcs11i.h"
 
@@ -394,6 +394,7 @@ service_method_open_session (GkdSecretService *self, DBusMessage *message)
 static DBusMessage*
 service_method_create_collection (GkdSecretService *self, DBusMessage *message)
 {
+	GckBuilder builder = GCK_BUILDER_INIT;
 	DBusMessageIter iter, array;
 	GckAttributes *attrs;
 	GkdSecretCreate *create;
@@ -408,10 +409,9 @@ service_method_create_collection (GkdSecretService *self, DBusMessage *message)
 		return NULL;
 	if (!dbus_message_iter_init (message, &iter))
 		g_return_val_if_reached (NULL);
-	attrs = gck_attributes_new ();
 	dbus_message_iter_recurse (&iter, &array);
-	if (!gkd_secret_property_parse_all (&array, SECRET_COLLECTION_INTERFACE, attrs)) {
-		gck_attributes_unref (attrs);
+	if (!gkd_secret_property_parse_all (&array, SECRET_COLLECTION_INTERFACE, &builder)) {
+		gck_builder_clear (&builder);
 		return dbus_message_new_error_printf (message, DBUS_ERROR_INVALID_ARGS,
 		                                      "Invalid properties");
 	}
@@ -424,13 +424,14 @@ service_method_create_collection (GkdSecretService *self, DBusMessage *message)
 		if (!alias[0]) {
 			alias = NULL;
 		} else if (!g_str_equal (alias, "default")) {
-			gck_attributes_unref (attrs);
+			gck_builder_clear (&builder);
 			return dbus_message_new_error (message, DBUS_ERROR_NOT_SUPPORTED,
 			                               "Only the 'default' alias is supported");
 		}
 	}
 
-	gck_attributes_add_boolean (attrs, CKA_TOKEN, TRUE);
+	gck_builder_add_boolean (&builder, CKA_TOKEN, TRUE);
+	attrs = gck_attributes_ref_sink (gck_builder_end (&builder));
 
 	/* Create the prompt object, for the password */
 	caller = dbus_message_get_sender (message);
@@ -663,11 +664,13 @@ service_method_set_alias (GkdSecretService *self, DBusMessage *message)
 static DBusMessage*
 service_method_create_with_master_password (GkdSecretService *self, DBusMessage *message)
 {
+	GckBuilder builder = GCK_BUILDER_INIT;
 	DBusError derr = DBUS_ERROR_INIT;
 	DBusMessageIter iter, array;
 	DBusMessage *reply = NULL;
 	GkdSecretSecret *secret = NULL;
 	GckAttributes *attrs = NULL;
+	GError *error = NULL;
 	gchar *path;
 
 	/* Parse the incoming message */
@@ -675,27 +678,27 @@ service_method_create_with_master_password (GkdSecretService *self, DBusMessage
 		return NULL;
 	if (!dbus_message_iter_init (message, &iter))
 		g_return_val_if_reached (NULL);
-	attrs = gck_attributes_new ();
 	dbus_message_iter_recurse (&iter, &array);
-	if (!gkd_secret_property_parse_all (&array, SECRET_COLLECTION_INTERFACE, attrs)) {
-		gck_attributes_unref (attrs);
+	if (!gkd_secret_property_parse_all (&array, SECRET_COLLECTION_INTERFACE, &builder)) {
+		gck_builder_clear (&builder);
 		return dbus_message_new_error (message, DBUS_ERROR_INVALID_ARGS,
 		                               "Invalid properties argument");
 	}
 	dbus_message_iter_next (&iter);
 	secret = gkd_secret_secret_parse (self, message, &iter, &derr);
 	if (secret == NULL) {
-		gck_attributes_unref (attrs);
+		gck_builder_clear (&builder);
 		return gkd_secret_error_to_reply (message, &derr);
 	}
 
-	gck_attributes_add_boolean (attrs, CKA_TOKEN, TRUE);
-	path = gkd_secret_create_with_secret (attrs, secret, &derr);
+	gck_builder_add_boolean (&builder, CKA_TOKEN, TRUE);
+	attrs = gck_attributes_ref_sink (gck_builder_end (&builder));
+	path = gkd_secret_create_with_secret (attrs, secret, &error);
 	gck_attributes_unref (attrs);
 	gkd_secret_secret_free (secret);
 
 	if (path == NULL)
-		return gkd_secret_error_to_reply (message, &derr);
+		return gkd_secret_propagate_error (message, "Couldn't create collection", error);
 
 	reply = dbus_message_new_method_return (message);
 	dbus_message_append_args (reply, DBUS_TYPE_OBJECT_PATH, &path, DBUS_TYPE_INVALID);
@@ -712,6 +715,7 @@ service_method_change_with_master_password (GkdSecretService *self, DBusMessage
 	GckObject *collection;
 	DBusMessageIter iter;
 	DBusMessage *reply;
+	GError *error = NULL;
 	const gchar *path;
 
 	/* Parse the incoming message */
@@ -742,12 +746,12 @@ service_method_change_with_master_password (GkdSecretService *self, DBusMessage
 		                                "The collection does not exist");
 
 	/* Success */
-	else if (gkd_secret_change_with_secrets (collection, original, master, &derr))
+	else if (gkd_secret_change_with_secrets (collection, NULL, original, master, &error))
 		reply = dbus_message_new_method_return (message);
 
 	/* Failure */
 	else
-		reply = gkd_secret_error_to_reply (message, &derr);
+		reply = gkd_secret_propagate_error (message, "Couldn't change collection password", error);
 
 	gkd_secret_secret_free (original);
 	gkd_secret_secret_free (master);
@@ -763,6 +767,7 @@ service_method_unlock_with_master_password (GkdSecretService *self, DBusMessage
 {
 	DBusError derr = DBUS_ERROR_INIT;
 	GkdSecretSecret *master;
+	GError *error = NULL;
 	GckObject *collection;
 	DBusMessageIter iter;
 	DBusMessage *reply;
@@ -790,12 +795,12 @@ service_method_unlock_with_master_password (GkdSecretService *self, DBusMessage
 		                                "The collection does not exist");
 
 	/* Success */
-	else if (gkd_secret_unlock_with_secret (collection, master, &derr))
+	else if (gkd_secret_unlock_with_secret (collection, master, &error))
 		reply = dbus_message_new_method_return (message);
 
 	/* Failure */
 	else
-		reply = gkd_secret_error_to_reply (message, &derr);
+		reply = gkd_secret_propagate_error (message, "Couldn't unlock collection", error);
 
 	gkd_secret_secret_free (master);
 
@@ -805,6 +810,52 @@ service_method_unlock_with_master_password (GkdSecretService *self, DBusMessage
 	return reply;
 }
 
+static void
+on_each_path_append_to_array (GkdSecretObjects *self,
+                              const gchar *path,
+                              GckObject *object,
+                              gpointer user_data)
+{
+	GPtrArray *array = user_data;
+	g_ptr_array_add (array, g_strdup (path));
+}
+
+static DBusMessage *
+service_introspect (GkdSecretService *self,
+                    DBusMessage *message)
+{
+	GPtrArray *names;
+	DBusMessage *reply;
+	ServiceClient *client;
+	const gchar *caller;
+	const gchar *path;
+	GHashTableIter iter;
+
+	names = g_ptr_array_new_with_free_func (g_free);
+	gkd_secret_objects_foreach_collection (self->objects, message,
+	                                       on_each_path_append_to_array,
+	                                       names);
+
+	/* Lookup all sessions and prompts for this client */
+	caller = dbus_message_get_sender (message);
+	if (caller != NULL) {
+		client = g_hash_table_lookup (self->clients, caller);
+		if (client != NULL) {
+			g_hash_table_iter_init (&iter, client->dispatch);
+			while (g_hash_table_iter_next (&iter, (gpointer *)&path, NULL))
+				g_ptr_array_add (names, g_strdup (path));
+		}
+	}
+
+	g_ptr_array_add (names, NULL);
+
+	reply = gkd_dbus_introspect_handle (message, gkd_secret_introspect_collection,
+	                                    (const gchar **)names->pdata);
+
+	g_ptr_array_unref (names);
+	return reply;
+}
+
 static DBusMessage*
 service_message_handler (GkdSecretService *self, DBusMessage *message)
 {
@@ -875,12 +926,31 @@ service_message_handler (GkdSecretService *self, DBusMessage *message)
 	else if (dbus_message_is_method_call (message, DBUS_INTERFACE_PROPERTIES, "GetAll"))
 		return service_property_getall (self, message);
 
+	/* org.freedesktop.DBus.Introspectable.Introspect() */
 	else if (dbus_message_has_interface (message, DBUS_INTERFACE_INTROSPECTABLE))
-		return gkd_dbus_introspect_handle (message, gkd_secret_introspect_service);
+		return service_introspect (self, message);
 
 	return NULL;
 }
 
+static gboolean
+root_dispatch_message (GkdSecretService *self,
+                       DBusMessage *message)
+{
+	DBusMessage *reply = NULL;
+
+	if (dbus_message_has_interface (message, DBUS_INTERFACE_INTROSPECTABLE))
+		reply = gkd_dbus_introspect_handle (message, gkd_secret_introspect_root, NULL);
+
+	if (reply != NULL) {
+		dbus_connection_send (self->connection, reply, NULL);
+		dbus_message_unref (reply);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
 static void
 service_dispatch_message (GkdSecretService *self, DBusMessage *message)
 {
@@ -987,6 +1057,11 @@ gkd_secret_service_filter_handler (DBusConnection *conn, DBusMessage *message, g
 
 	/* Dispatch any method call on our interfaces, for our objects */
 	case DBUS_MESSAGE_TYPE_METHOD_CALL:
+		if (path != NULL && g_str_equal (path, "/")) {
+			if (root_dispatch_message (self, message))
+				return DBUS_HANDLER_RESULT_HANDLED;
+		}
+
 		if (object_path_has_prefix (path, SECRET_SERVICE_PATH)) {
 			interface = dbus_message_get_interface (message);
 			if (interface == NULL ||
@@ -1345,12 +1420,19 @@ gkd_secret_service_get_alias (GkdSecretService *self, const gchar *alias)
 	const gchar *identifier;
 
 	g_return_val_if_fail (GKD_SECRET_IS_SERVICE (self), NULL);
-	g_return_val_if_fail (alias, NULL);
+	g_return_val_if_fail (alias != NULL, NULL);
 
 	identifier =  g_hash_table_lookup (self->aliases, alias);
 	if (!identifier && g_str_equal (alias, "default")) {
 		update_default (self, TRUE);
 		identifier = g_hash_table_lookup (self->aliases, alias);
+
+		/* Default to to 'login' if no default keyring */
+		if (identifier == NULL) {
+			identifier = "login";
+			g_hash_table_replace (self->aliases, g_strdup (alias),
+			                      g_strdup (identifier));
+		}
 	}
 	return identifier;
 }
