@@ -362,6 +362,7 @@ open_password_prompt (GckSession *session,
 	gcr_prompt_set_description (prompt, description);
 
 	gcr_prompt_set_password_new (prompt, confirm);
+	gcr_prompt_set_continue_label (prompt, _("Unlock"));
 
 	if (errmsg)
 		gcr_prompt_set_warning (prompt, errmsg);
@@ -422,7 +423,7 @@ do_get_password (GckSession *session, const gchar *keyid, const gchar *errmsg,
 	if (prompt != NULL) {
 		password = egg_secure_strdup (gcr_prompt_password (prompt, NULL, &error));
 		if (password == NULL) {
-			if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+			if (error && !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
 				g_warning ("couldn't prompt for password: %s", egg_error_message (error));
 			g_clear_error (&error);
 		}
