@@ -48,6 +48,7 @@
 #include "config.h"
 
 #include "egg-asn1x.h"
+#include "egg-timegm.h"
 
 #include <libtasn1.h>
 
@@ -158,7 +159,7 @@ struct _Abits {
 /* Forward Declarations */
 static gboolean anode_decode_anything (GNode*, Atlv*);
 static gboolean anode_decode_anything_for_flags (GNode *, Atlv*, gint);
-static gboolean anode_validate_anything (GNode*);
+static gboolean anode_validate_anything (GNode*, gboolean);
 static gboolean anode_encode_prepare (GNode*, gboolean want);
 
 static gint
@@ -1139,14 +1140,12 @@ anode_decode_anything (GNode *node, Atlv *tlv)
 }
 
 gboolean
-egg_asn1x_decode (GNode *asn, gconstpointer data, gsize n_data)
+egg_asn1x_decode_no_validate (GNode *asn,
+                              gconstpointer data,
+                              gsize n_data)
 {
 	Atlv tlv;
 
-	g_return_val_if_fail (asn, FALSE);
-	g_return_val_if_fail (data, FALSE);
-	g_return_val_if_fail (n_data, FALSE);
-
 	egg_asn1x_clear (asn);
 
 	if (!anode_decode_tlv_for_data (data, (const guchar*)data + n_data, &tlv))
@@ -1158,7 +1157,25 @@ egg_asn1x_decode (GNode *asn, gconstpointer data, gsize n_data)
 	if (tlv.end - tlv.buf != n_data)
 		return FALSE;
 
-	return egg_asn1x_validate (asn);
+	return TRUE;
+}
+
+gboolean
+egg_asn1x_decode (GNode *asn,
+                  gconstpointer data,
+                  gsize n_data)
+{
+	gboolean ret;
+
+	g_return_val_if_fail (asn, FALSE);
+	g_return_val_if_fail (data, FALSE);
+	g_return_val_if_fail (n_data, FALSE);
+
+	ret = egg_asn1x_decode_no_validate (asn, data, n_data);
+	if (!ret)
+		return ret;
+
+	return egg_asn1x_validate (asn, TRUE);
 }
 
 /* -----------------------------------------------------------------------------------
@@ -1474,6 +1491,33 @@ anode_encoder_simple (gpointer user_data, guchar *data, gsize n_data)
 	return TRUE;
 }
 
+static gboolean
+anode_encoder_unsigned (gpointer user_data,
+                        guchar *data,
+                        gsize n_data)
+{
+	gboolean sign;
+	gchar *p;
+
+	/*
+	 * If top bit is set, the result would be negative in two's complement
+	 * but since we want an unsigned integer, add a zero byte. That zero
+	 * byte is already calculated into n_data, see egg_asn1x_set_integer_as_usg
+	 */
+
+	p = user_data;
+	sign = !!(p[0] & 0x80);
+	if (sign) {
+		g_assert (n_data > 1);
+		data[0] = 0;
+		data++;
+		n_data--;
+	}
+
+	memcpy (data, p, n_data);
+	return TRUE;
+}
+
 static gboolean
 anode_encoder_structured (gpointer user_data, guchar *data, gsize n_data)
 {
@@ -1705,7 +1749,7 @@ egg_asn1x_encode (GNode *asn, EggAllocator allocator, gsize *n_data)
 		return NULL;
 
 	if (anode_encode_build (asn, data, length) &&
-	    anode_validate_anything (asn)) {
+	    anode_validate_anything (asn, TRUE)) {
 		anode_encode_commit (asn);
 		*n_data = length;
 		return data;
@@ -1762,36 +1806,6 @@ two_to_four_digit_year (int year)
 		return century + year;
 }
 
-#ifndef HAVE_TIMEGM
-time_t timegm(struct tm *t)
-{
-	time_t tl, tb;
-	struct tm *tg;
-
-	tl = mktime (t);
-	if (tl == -1)
-	{
-		t->tm_hour--;
-		tl = mktime (t);
-		if (tl == -1)
-			return -1; /* can't deal with output from strptime */
-		tl += 3600;
-	}
-	tg = gmtime (&tl);
-	tg->tm_isdst = 0;
-	tb = mktime (tg);
-	if (tb == -1)
-	{
-		tg->tm_hour--;
-		tb = mktime (tg);
-		if (tb == -1)
-			return -1; /* can't deal with output from gmtime */
-		tb += 3600;
-	}
-	return (tl - (tb - tl));
-}
-#endif // NOT_HAVE_TIMEGM
-
 static gboolean
 parse_utc_time (const gchar *time, gsize n_time,
                 struct tm* when, gint *offset)
@@ -2063,6 +2077,8 @@ anode_write_integer_ulong (gulong value, guchar *data, gsize *n_data)
 {
 	guchar buf[sizeof (gulong)];
 	gint bytes, i, off;
+	guchar *at;
+	gboolean sign;
 
 	for (i = 0; i < sizeof (gulong); ++i) {
 		off = sizeof (gulong) - (i + 1);
@@ -2077,11 +2093,20 @@ anode_write_integer_ulong (gulong value, guchar *data, gsize *n_data)
 	if (bytes == 0)
 		bytes = 1;
 
+	/* If the first byte would make this negative, then add a zero */
+	at = buf + (sizeof (gulong) - bytes);
+	sign = !!(at[0] & 0x80);
+
 	if (data) {
-		g_assert (*n_data >= bytes);
-		memcpy (data, buf + (sizeof (gulong) - bytes), bytes);
+		g_assert (*n_data >= bytes + 1);
+		if (sign) {
+			data[0] = 0;
+			data++;
+		}
+		memcpy (data, at, bytes);
 	}
-	*n_data = bytes;
+
+	*n_data = bytes + (sign ? 1 : 0);
 	return TRUE;
 }
 
@@ -2523,7 +2548,7 @@ egg_asn1x_set_enumerated (GNode *node, GQuark value)
 	val = anode_def_value_as_ulong (opt);
 	g_return_val_if_fail (val != G_MAXULONG, FALSE);
 
-	n_data = sizeof (gulong);
+	n_data = sizeof (gulong) + 1;
 	data = g_malloc0 (n_data);
 	if (!anode_write_integer_ulong (val, data, &n_data))
 		return FALSE;
@@ -2582,7 +2607,7 @@ egg_asn1x_set_integer_as_ulong (GNode *node, gulong value)
 
 	/* TODO: Handle default values */
 
-	n_data = sizeof (gulong);
+	n_data = sizeof (gulong) + 1;
 	data = g_malloc0 (n_data);
 	if (!anode_write_integer_ulong (value, data, &n_data))
 		return FALSE;
@@ -2590,30 +2615,54 @@ egg_asn1x_set_integer_as_ulong (GNode *node, gulong value)
 	return TRUE;
 }
 
-gpointer
-egg_asn1x_get_integer_as_raw (GNode *node, EggAllocator allocator, gsize *n_data)
+gconstpointer
+egg_asn1x_get_integer_as_raw (GNode *node, gsize *n_data)
 {
 	Atlv *tlv;
-	gpointer data;
 
 	g_return_val_if_fail (node, FALSE);
 	g_return_val_if_fail (n_data, FALSE);
 	g_return_val_if_fail (anode_def_type (node) == TYPE_INTEGER, FALSE);
 
-	if (!allocator)
-		allocator = g_realloc;
-
 	tlv = anode_get_tlv_data (node);
 	if (tlv == NULL || tlv->buf == NULL)
 		return NULL;
 
-	data = (allocator) (NULL, tlv->len);
-	if (data == NULL)
-		return NULL;
-
-	memcpy (data, tlv->buf + tlv->off, tlv->len);
 	*n_data = tlv->len;
-	return data;
+	return tlv->buf + tlv->off;
+}
+
+gconstpointer
+egg_asn1x_get_integer_as_usg (GNode *node,
+                              gsize *n_data)
+{
+	const guchar *p;
+	gboolean sign;
+	gsize len;
+
+	g_return_val_if_fail (node, FALSE);
+	g_return_val_if_fail (n_data, FALSE);
+	g_return_val_if_fail (anode_def_type (node) == TYPE_INTEGER, FALSE);
+
+	p = egg_asn1x_get_integer_as_raw (node, &len);
+	sign = !!(p[0] & 0x80);
+	if (sign) {
+		g_warning ("invalid two's complement integer is negative, but expected unsigned");
+		return NULL;
+	}
+
+	*n_data = len;
+
+	/* Strip off the extra zero byte that was preventing it from being negative */
+	if (p[0] == 0 && len > 1) {
+		sign = !!(p[1] & 0x80);
+		if (sign) {
+			p++;
+			*n_data = len - 1;
+		}
+	}
+
+	return p;
 }
 
 gboolean
@@ -2639,6 +2688,37 @@ egg_asn1x_set_integer_as_raw (GNode *node, gconstpointer data, gsize n_data, GDe
 	return TRUE;
 }
 
+gboolean
+egg_asn1x_set_integer_as_usg (GNode *node,
+                              gconstpointer data,
+                              gsize n_data,
+                              GDestroyNotify destroy)
+{
+	gboolean sign;
+	guchar *p;
+
+	g_return_val_if_fail (node, FALSE);
+	g_return_val_if_fail (data, FALSE);
+	g_return_val_if_fail (n_data > 0, FALSE);
+	g_return_val_if_fail (anode_def_type (node) == TYPE_INTEGER, FALSE);
+
+	/* Make sure the integer is properly encoded in twos complement*/
+	p = (guchar*)data;
+	sign = !!(p[0] & 0x80);
+
+	/*
+	 * If in two's complement this would be negative, add a zero byte so
+	 * that it isn't. Here we just note that the result will be one byte
+	 * longer. In anode_encoder_unsigned we actually add the zero byte.
+	 */
+	if (sign)
+		n_data += 1;
+
+	anode_encode_tlv_and_enc (node, n_data, anode_encoder_unsigned,
+	                          (gpointer)data, destroy);
+	return TRUE;
+}
+
 gconstpointer
 egg_asn1x_get_raw_element (GNode *node, gsize *n_element)
 {
@@ -3334,7 +3414,8 @@ anode_validate_time (GNode *node, Atlv *tlv)
 }
 
 static gboolean
-anode_validate_choice (GNode *node)
+anode_validate_choice (GNode *node,
+                       gboolean strict)
 {
 	GNode *child, *choice;
 	Anode *an;
@@ -3344,7 +3425,7 @@ anode_validate_choice (GNode *node)
 	if (!choice)
 		return anode_failure (node, "one choice must be set");
 
-	if (!anode_validate_anything (choice))
+	if (!anode_validate_anything (choice, strict))
 		return FALSE;
 
 	for (child = node->children; child; child = child->next) {
@@ -3359,7 +3440,8 @@ anode_validate_choice (GNode *node)
 }
 
 static gboolean
-anode_validate_sequence_or_set (GNode *node)
+anode_validate_sequence_or_set (GNode *node,
+                                gboolean strict)
 {
 	GNode *child;
 	gulong tag = 0;
@@ -3371,7 +3453,7 @@ anode_validate_sequence_or_set (GNode *node)
 
 	/* All of the children must validate properly */
 	for (child = node->children; child; child = child->next) {
-		if (!anode_validate_anything (child))
+		if (!anode_validate_anything (child, strict))
 			return FALSE;
 
 		/* Tags must be in ascending order */
@@ -3388,7 +3470,8 @@ anode_validate_sequence_or_set (GNode *node)
 }
 
 static gboolean
-anode_validate_sequence_or_set_of (GNode *node)
+anode_validate_sequence_or_set_of (GNode *node,
+                                   gboolean strict)
 {
 	GNode *child;
 	Atlv *tlv, *ptlv;
@@ -3406,7 +3489,7 @@ anode_validate_sequence_or_set_of (GNode *node)
 	for (child = node->children; child; child = child->next) {
 		tlv = anode_get_tlv_data (child);
 		if (tlv) {
-			if (!anode_validate_anything (child))
+			if (!anode_validate_anything (child, strict))
 				return FALSE;
 
 			/* Tag must have same tag as top */
@@ -3416,7 +3499,8 @@ anode_validate_sequence_or_set_of (GNode *node)
 				return anode_failure (node, "invalid mismatched content");
 
 			/* Set of must be in ascending order */
-			if (type == TYPE_SET_OF && ptlv && compare_tlvs (ptlv, tlv) > 0)
+			if (strict && type == TYPE_SET_OF &&
+			    ptlv != NULL && compare_tlvs (ptlv, tlv) > 0)
 				return anode_failure (node, "content must be in ascending order");
 			ptlv = tlv;
 			++count;
@@ -3427,7 +3511,8 @@ anode_validate_sequence_or_set_of (GNode *node)
 }
 
 static gboolean
-anode_validate_anything (GNode *node)
+anode_validate_anything (GNode *node,
+                         gboolean strict)
 {
 	Atlv *tlv;
 	gint type;
@@ -3471,16 +3556,16 @@ anode_validate_anything (GNode *node)
 	case TYPE_ANY:
 		return TRUE;
 	case TYPE_CHOICE:
-		return anode_validate_choice (node);
+		return anode_validate_choice (node, strict);
 
 	/* Structured types */
 	case TYPE_SEQUENCE:
 	case TYPE_SET:
-		return anode_validate_sequence_or_set (node);
+		return anode_validate_sequence_or_set (node, strict);
 
 	case TYPE_SEQUENCE_OF:
 	case TYPE_SET_OF:
-		return anode_validate_sequence_or_set_of (node);
+		return anode_validate_sequence_or_set_of (node, strict);
 
 	default:
 		g_return_val_if_reached (FALSE);
@@ -3488,10 +3573,11 @@ anode_validate_anything (GNode *node)
 }
 
 gboolean
-egg_asn1x_validate (GNode *asn)
+egg_asn1x_validate (GNode *asn,
+                    gboolean strict)
 {
 	g_return_val_if_fail (asn, FALSE);
-	return anode_validate_anything (asn);
+	return anode_validate_anything (asn, strict);
 }
 
 /* -----------------------------------------------------------------------------------
